#!python3
# -*- coding: utf-8 -*-
"""
File format base class
"""
import argparse
import shlex
from abc import ABC, abstractmethod
from typing import Union, Dict
import io

import pandas as pd

from ckanapi_harvesters.auxiliary.ckan_model import CkanField, CkanResourceInfo
from ckanapi_harvesters.auxiliary.list_records import ListRecords
from ckanapi_harvesters.auxiliary.ckan_auxiliary import import_args_kwargs_dict


class FileFormatABC(ABC):
    default_read_chunksize: int = 10000
    default_read_kwargs:dict = {}
    default_write_kwargs:dict = {}

    def __init__(self, options_string: str=None, *, read_kwargs:dict=None, write_kwargs:dict=None):
        if read_kwargs is None: read_kwargs = self.default_read_kwargs
        if write_kwargs is None: write_kwargs = self.default_write_kwargs
        self.options_string:Union[str,None] = options_string
        self.allow_chunks:bool = True
        self.chunk_size:int = self.default_read_chunksize
        self.read_kwargs:dict = read_kwargs
        self.write_kwargs:dict = write_kwargs
        self.extra_args: Union[list,None] = None
        self.resource_attributes_from_file: Union[CkanResourceInfo,None] = None  # destination for resource attributes extracted from data source, if any
        self._apply_options_string()
        self.allow_chunks = self.allow_chunks and self.read_by_chunks_allowed()  # override

    def _get_read_kwargs(self, allow_chunks:bool=True) -> dict:
        kwargs = self.default_read_kwargs.copy()
        kwargs.update(self.read_kwargs)
        return kwargs

    def _get_write_kwargs(self) -> dict:
        kwargs = self.default_write_kwargs.copy()
        kwargs.update(self.write_kwargs)
        return kwargs

    @staticmethod
    def _setup_cli_parser(parser: argparse.ArgumentParser = None) -> argparse.ArgumentParser:
        if parser is None:
            parser = argparse.ArgumentParser(description="File format reader arguments", add_help=False,
                                             epilog=
                                             "Examples: \n"
                                             "- Changing chunk size: --chunk-size 10000 \n"
                                             "- Disabling reading files by chunks: --no-chunks \n"
                                             "- Additional arguments for pandas.read_csv for a CSV file: --read-kwargs compression=gzip header=10")
        parser.add_argument("--chunk-size", type=int,
                            help="Chunk size for reading files by chunks (number of records).\n"
                            "The number of lines sent per request is the minimum of chunk size and CKAN parameter ckan.params.default_limit_write\n"
                            "Enabling this option activates reading by chunks (if supported by the file format)")
        parser.add_argument("--no-chunks",
                            help="Option to disabling reading files by chunks", action="store_true", default=None)
        parser.add_argument("--allow-chunks",
                            help="Option to enable reading files by chunks (useful for file formats not enabling this feature by default)", action="store_true", default=None)
        parser.add_argument("--read-kwargs", nargs="*",
                            help="Keyword arguments for the read function in key=value format")
        parser.add_argument("--write-kwargs", nargs="*",
                            help="Keyword arguments for the write function in key=value format")
        return parser

    def print_help_cli(self, display:bool=True) -> str:
        parser = self._setup_cli_parser()
        if display:
            parser.print_help()
        with io.StringIO() as buffer:
            parser.print_help(buffer)
            return buffer.getvalue()

    def _apply_arguments(self, args: argparse.Namespace, extra_args: list):
        if args.no_chunks is not None:
            self.allow_chunks = not args.no_chunks
        if args.allow_chunks is not None:
            self.allow_chunks = args.allow_chunks
        if args.chunk_size is not None:
            self.chunk_size = args.chunk_size
            self.allow_chunks = True
        self.extra_args = extra_args
        self.read_kwargs.update(import_args_kwargs_dict(args.read_kwargs))
        self.write_kwargs.update(import_args_kwargs_dict(args.write_kwargs))

    def _apply_options_string(self, options_string:str=None, *, parser: argparse.ArgumentParser = None):
        if options_string is None:
            options_string = self.options_string
        if options_string is None:
            options_string = ""
        parser = self._setup_cli_parser(parser)
        args, extra_args = parser.parse_known_args(shlex.split(options_string))
        self._apply_arguments(args, extra_args)

    # read -------------------
    @abstractmethod
    def read_by_chunks_allowed(self) -> bool:
        raise NotImplementedError()

    def read_by_chunks_enabled(self, allow_chunks:bool=True) -> bool:
        return self.read_by_chunks_allowed() and self.allow_chunks and allow_chunks

    @abstractmethod
    def read_file(self, file_path: str, fields: Union[Dict[str, CkanField],None], allow_chunks:bool=True) -> Union[pd.DataFrame, ListRecords]:
        """
        Read a file from the file system, either fully (returning DataFrame or ListRecords) or by chunks (Iterator over a number of records).
        """
        raise NotImplementedError()

    @abstractmethod
    def read_buffer_full(self, buffer: io.IOBase, fields: Union[Dict[str, CkanField],None]) -> Union[pd.DataFrame, ListRecords]:
        """
        Read a file from memory, as a DataFrame or ListRecords. This function reads entirely the file.
        """
        raise NotImplementedError()

    # write -------------------
    @abstractmethod
    def write_file(self, df: Union[pd.DataFrame, ListRecords], file_path: str, fields: Union[Dict[str, CkanField],None]) -> None:
        """
        Write a DataFrame or ListRecords to a file.
        """
        raise NotImplementedError()

    @abstractmethod
    def append_allowed(self) -> bool:
        """
        This function announces that the file format is allowed to be written in append mode, and that the function append_file is implemented.
        """
        return False

    def append_file(self, df: Union[pd.DataFrame, ListRecords], file_path: str, fields: Union[Dict[str, CkanField],None]) -> None:
        """
        Write a DataFrame or ListRecords to a file, appending it to the end of the file.
        """
        raise NotImplementedError()

    @abstractmethod
    def write_in_memory(self, df: Union[pd.DataFrame, ListRecords], fields: Union[Dict[str, CkanField],None]) -> bytes:
        """
        This function writes a DataFrame or ListRecords to a file in memory.
        """
        raise NotImplementedError()

    def append_in_memory(self, stream: bytes, df: Union[pd.DataFrame, ListRecords], fields: Union[Dict[str, CkanField],None]) -> bytes:
        """
        This function writes a DataFrame or ListRecords to a file in memory, appending its to the end of the buffer.
        """
        raise NotImplementedError()

    # misc -------------------
    @abstractmethod
    def copy(self, dest=None):
        dest.options_string = self.options_string
        dest.allow_chunks = self.allow_chunks
        dest.chunk_size = self.chunk_size
        dest.read_kwargs = self.read_kwargs
        dest.write_kwargs = self.write_kwargs
        return dest

    def __copy__(self):
        return self.copy()




#!python3
# -*- coding: utf-8 -*-
"""
Code to upload metadata to the CKAN server to create/update an existing package
The metadata is defined by the user in an Excel worksheet
This file implements the basic resources. See builder_datastore for specific functions to initiate datastores.
"""
from concurrent.futures import ThreadPoolExecutor
import threading
from threading import current_thread, Semaphore
from typing import Any, Generator, List, Dict
from abc import ABC, abstractmethod
from warnings import warn


from ckanapi_harvesters.auxiliary.ckan_errors import UnexpectedError
from ckanapi_harvesters.ckan_api import CkanApi
from ckanapi_harvesters.auxiliary.ckan_auxiliary import FileChunkDataFrame
from ckanapi_harvesters.auxiliary.ckan_progress_callbacks import CkanProgressCallback, CkanCallbackLevel
from ckanapi_harvesters.builder.builder_aux import positive_end_index

multi_file_exclude_other_files:bool = True


class BuilderMultiABC(ABC):
    def __init__(self):
        self.progress_callback = CkanProgressCallback()
        self.stop_event = threading.Event()
        self.thread_ckan: Dict[str, CkanApi] = {}
        self.enable_multi_threaded_upload:bool = True
        self.enable_multi_threaded_download:bool = True
        self.file_semaphore = Semaphore()
        # from Resource (for code validation)
        self.name:str = ""
        self.enable_download:bool = True
        self.read_line_counter:int = 0

    def copy(self, *, dest=None):
        dest.progress_callback = self.progress_callback.copy()
        dest.enable_multi_threaded_upload = self.enable_multi_threaded_upload
        dest.enable_multi_threaded_download = self.enable_multi_threaded_download
        # do not copy stop_event
        return dest

    def _call_progress_callback(self, position:int, total:int, *, info:Any=None, context:str=None,
                                file_index:int=0, file_count:int=None, lines_chunk:int=None, total_lines_read:int=None,
                                canceled_request: bool=False, end_message: bool=False, level:int=0) -> None:
        """
        Progress callback function. Use to implement a progress indication for the user.

        :param position: the position within the resource (usually, in bytes or line count)
        :param total: the total size of the resource
        :param info: an object from which more information can be extracted, typically, the DataFrame itself, with an indication of the data origin.
        :param context: the context of the call (ckan instance, upload/download, single/multi-threaded)
        :param file_index: the index of the file in the list
        :param file_count: the number of files in the list
        :param lines_chunk: the number of lines in the chunk currently being processed
        :param total_lines_read: the total number of lines read, including the current chunk
        :param canceled_request: this callback is also called when a line is ignored
        :param end_message: boolean indicating of the work in progress
        :param level: the level of the progress callback (1: package/dataset, 2: resource builder, 3: used for multi-file resources)
        """
        if self.progress_callback is not None:
            if end_message:
                position = total
                file_index = file_count
            self.progress_callback(position, total, info=info, context=context,
                                   file_index=file_index, file_count=file_count,
                                   lines_chunk=lines_chunk, total_lines_read=total_lines_read,
                                   canceled_upload=canceled_request, end_message=end_message,
                                   **self.progress_callback_kwargs)

    def _prepare_for_multithreading(self, ckan: CkanApi):
        self.stop_event.clear()
        self.thread_ckan.clear()

    def _init_thread(self, ckan: CkanApi):
        thread_name = current_thread().name
        ckan_thread = ckan.copy(new_identifier=thread_name)
        ckan_thread.prepare_for_multithreading(True)  # prepare CKAN object for multi-threading
        self.thread_ckan[thread_name] = ckan_thread

    def _terminate_thread(self):
        for ckan in self.thread_ckan.values():
            ckan.disconnect()
        self.thread_ckan.clear()


    ## upload -----------------------------------------------------------------
    @abstractmethod
    def init_local_files_list(self, resources_base_dir:str, ckan:CkanApi, cancel_if_present:bool=True, **kwargs) -> List[str]:
        """
        Behavior to list parts of an upload.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_local_file_offset(self, file_chunk: FileChunkDataFrame) -> int:
        """
        Get the position of the current data in the overall upload.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_local_file_total_size(self) -> int:
        """
        Get the overall size of the upload, normally in bytes or line count.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_local_file_count(self) -> int:
        """
        Get the number of parts of the upload.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_local_df_chunk_generator(self, resources_base_dir:str, ckan:CkanApi, **kwargs) -> Generator[FileChunkDataFrame, None, None]:
        """
        Returns an iterator over the data to upload and a position in the current file.
        """
        raise NotImplementedError()

    def upload_request_final(self, ckan:CkanApi, *, force:bool=False) -> None:
        raise UnexpectedError("Should not reach this function")

    def _unit_upload_apply(self, *, ckan:CkanApi, file_chunk: FileChunkDataFrame,
                           upload_alter:bool=True, overall_chunk_index:int, file_count: int, start_index:int, end_index:int, **kwargs) -> Any:
        """
        Unitary function deciding whether to perform upload and making the steps for the upload.
        """
        raise NotImplementedError()

    def upload_request_full(self, ckan:CkanApi, resources_base_dir:str, *,
                            threads:int=1, external_stop_event=None, from_line_count:bool=False,
                            allow_chunks:bool=True,
                            start_index:int=0, end_index:int=None, **kwargs) -> None:
        """
        Perform all the upload requests.

        :param ckan:
        :param resources_base_dir:
        :param threads:
        :param external_stop_event:
        :param only_missing:
        :param start_index:
        :param end_index:
        :return:
        """
        if threads < 0:
            # cancel large uploads in this case
            return None
        elif (threads is None or threads > 1) and self.enable_multi_threaded_upload:
            return self.upload_request_full_multi_threaded(ckan=ckan, resources_base_dir=resources_base_dir,
                                                           threads=threads, external_stop_event=external_stop_event,
                                                           start_index=start_index, end_index=end_index, **kwargs)
        else:
            self.init_local_files_list(resources_base_dir=resources_base_dir, ckan=ckan, cancel_if_present=True, **kwargs)
            if ckan.params.verbose_extra:
                print(f"Launching single-threaded upload of multi-file resource {self.name}")
            total = self.get_local_file_count()
            end_index = positive_end_index(end_index, total)
            self.read_line_counter = 0
            self.progress_callback.start_task(self.get_local_file_total_size(), file_count=total,
                                              context=f"{ckan.identifier} single-thread upload", level=CkanCallbackLevel.ResourceChunks)
            for overall_chunk_index, file_chunk in enumerate(self.get_local_df_chunk_generator(resources_base_dir=resources_base_dir, ckan=ckan, allow_chunks=allow_chunks, **kwargs)):
                if external_stop_event is not None and external_stop_event.is_set():
                    print(f"{ckan.identifier} Interrupted")
                    return
                self._unit_upload_apply(ckan=ckan, file_chunk=file_chunk, overall_chunk_index=overall_chunk_index,
                                        start_index=start_index, end_index=end_index, file_count=total, **kwargs)
            self.progress_callback.end_task(self.get_local_file_total_size(), file_count=total, total_lines_read=self.read_line_counter,
                                         context=f"{ckan.identifier} single-thread upload", level=CkanCallbackLevel.ResourceChunks)
            # at last, apply final actions:
            self.upload_request_final(ckan)

    def upload_request_graceful(self, ckan:CkanApi, file_chunk: FileChunkDataFrame, *, overall_chunk_index:int,
                                external_stop_event=None,
                                start_index:int=0, end_index:int=None, **kwargs) -> None:
        """
        Calls upload_file with checks specific to multi-threading.

        :return:
        """
        # ckan.session_reset()
        # ckan.identifier = current_thread().name
        ckan = self.thread_ckan[current_thread().name]
        total = self.get_local_file_count()
        end_index = positive_end_index(end_index, total)
        if self.stop_event.is_set():
            return
        if external_stop_event is not None and external_stop_event.is_set():
            print(f"{ckan.identifier} Interrupted")
            return
        try:
            self._unit_upload_apply(ckan=ckan, file_chunk=file_chunk, overall_chunk_index=overall_chunk_index,
                                    start_index=start_index, end_index=end_index, file_count=total, **kwargs)
        except Exception as e:
            self.stop_event.set()  # Ensure all threads stop
            if ckan.params.verbose_extra:
                print(f"Stopping all threads because an exception occurred in thread: {e}")
            raise e from e

    def upload_request_full_multi_threaded(self, ckan:CkanApi, resources_base_dir:str,
                                           threads:int=1, external_stop_event=None,
                                           allow_chunks:bool=True,
                                           start_index:int=0, end_index:int=None, **kwargs):
        """
        Multi-threaded implementation of upload_request_full, using ThreadPoolExecutor.
        """
        self.init_local_files_list(resources_base_dir=resources_base_dir, ckan=ckan, cancel_if_present=True, **kwargs)
        self._prepare_for_multithreading(ckan)
        self.progress_callback.start_task(self.get_local_file_total_size(), file_count=self.get_local_file_count(),
                                          context=f"{ckan.identifier} multi-thread upload", level=CkanCallbackLevel.ResourceChunks)
        try:
            with ThreadPoolExecutor(max_workers=threads, initializer=self._init_thread, initargs=(ckan,)) as executor:
                if ckan.params.verbose_extra:
                    print(f"Launching multi-threaded upload of multi-file resource {self.name}")
                futures = [executor.submit(self.upload_request_graceful, ckan=ckan, file_chunk=file_chunk, overall_chunk_index=overall_chunk_index,
                                           start_index=start_index, end_index=end_index, external_stop_event=external_stop_event,
                                           **kwargs)
                           for overall_chunk_index, file_chunk in enumerate(self.get_local_df_chunk_generator(resources_base_dir=resources_base_dir, ckan=ckan, allow_chunks=allow_chunks, **kwargs))]
                for future in futures:
                    future.result()  # This will propagate the exception
            self.progress_callback.end_task(self.get_local_file_total_size(), file_count=self.get_local_file_count(),
                                         context=f"{ckan.identifier} multi-thread upload", level=CkanCallbackLevel.ResourceChunks)
        except Exception as e:
            self.stop_event.set()  # Ensure all threads stop
            if ckan.params.verbose_extra:
                print(f"Stopping all threads because an exception occurred: {e}")
            raise e from e
        finally:
            # self.stop_event.set()  # Ensure all threads stop
            if ckan.params.verbose_extra:
                print("End of multi-threaded upload...")
        # at last, apply final actions:
        self._terminate_thread()
        self.upload_request_final(ckan)

    ## download -------------------------------------------------------------
    @abstractmethod
    def init_download_file_query_list(self, ckan: CkanApi, out_dir: str, cancel_if_present: bool = True, **kwargs) -> List[Any]:
        """
        Determine the list of queries to download to reconstruct the uploaded parts.
        By default, the unique combinations of the first columns of the primary key are used.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_file_query_generator(self) -> Generator[Any, Any, None]:
        """
        Returns an iterator on all the file_queries.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_file_query_count(self) -> int:
        """
        Returns the total number of file_queries.
        """
        raise NotImplementedError()

    @abstractmethod
    def download_file_query_item(self, ckan: CkanApi, out_dir: str, file_query_item: Any) -> Any:
        """
        Download the file_query item with the its arguments
        """
        raise NotImplementedError()

    @abstractmethod
    def download_request_generator(self, ckan: CkanApi, out_dir: str) -> Generator[Any, Any, None]:
        """
        Generator to apply treatments after each request (single-threaded).

        :param ckan:
        :param out_dir:
        :return:
        """
        raise NotImplementedError()

    @abstractmethod
    def _unit_download_apply(self, ckan:CkanApi, file_query_item:Any, out_dir:str,
                             index:int, start_index:int, end_index:int, total:int,
                             **kwargs) -> Any:
        """
        Unitary function deciding whether to perform download and making the steps for the request.
        """
        raise NotImplementedError()

    def download_request_full(self, ckan: CkanApi, out_dir: str, threads:int=1, external_stop_event=None,
                              start_index:int=0, end_index:int=None, force:bool=False, **kwargs) -> None:

        if (not self.enable_download) and (not force):
            msg = f"Did not download resource {self.name} because download was disabled."
            warn(msg)
            return None
        if threads < 0:
            # do not download large datasets in this case
            return None
        elif (threads is None or threads > 1) and self.enable_multi_threaded_download:
            return self.download_request_full_multi_threaded(ckan=ckan, out_dir=out_dir,
                                                             threads=threads, external_stop_event=external_stop_event,
                                                             start_index=start_index, end_index=end_index, **kwargs)
        else:
            self.init_download_file_query_list(ckan=ckan, out_dir=out_dir, cancel_if_present=True, **kwargs)
            if ckan.params.verbose_extra:
                print(f"Launching single-threaded download of multi-file resource {self.name}")
            total = self.get_file_query_count()
            end_index = positive_end_index(end_index, total)
            self.read_line_counter = 0
            self.progress_callback.start_task(total, file_count=total, context=f"{ckan.identifier} single-thread download", level=CkanCallbackLevel.ResourceChunks)
            for index, file_query_item in enumerate(self.get_file_query_generator()):
                if external_stop_event is not None and external_stop_event.is_set():
                    print(f"{ckan.identifier} Interrupted")
                    return
                self._unit_download_apply(ckan=ckan, file_query_item=file_query_item, out_dir=out_dir,
                                          index=index, start_index=start_index, end_index=end_index, total=total, **kwargs)
            self.progress_callback.end_task(total, file_count=total, context=f"{ckan.identifier} single-thread download", level=CkanCallbackLevel.ResourceChunks)

    def download_file_query_item_graceful(self, ckan: CkanApi, out_dir: str, file_query_item: Any, index:int,
                                          external_stop_event=None, start_index:int=0, end_index:int=None, **kwargs) -> None:
        """
        Implementation of download_file_query_item with checks for a multi-threaded download.
        """
        # ckan.session_reset()
        # ckan.identifier = current_thread().name
        ckan = self.thread_ckan[current_thread().name]
        total = self.get_file_query_count()
        end_index = positive_end_index(end_index, total)
        if self.stop_event.is_set():
            return
        if external_stop_event is not None and external_stop_event.is_set():
            print(f"{ckan.identifier} Interrupted")
            return
        try:
            self._unit_download_apply(ckan=ckan, file_query_item=file_query_item, out_dir=out_dir,
                                      index=index, start_index=start_index, end_index=end_index, total=total, **kwargs)
        except Exception as e:
            self.stop_event.set()  # Ensure all threads stop
            if ckan.params.verbose_extra:
                print(f"Stopping all threads because an exception occurred in thread: {e}")
            raise e from e

    def download_request_full_multi_threaded(self, ckan: CkanApi, out_dir: str,
                                             threads: int = None, external_stop_event=None,
                                             start_index:int=0, end_index:int=-1, **kwargs) -> None:
        """
        Multi-threaded implementation of download_request_full using ThreadPoolExecutor.
        """
        self.init_download_file_query_list(ckan=ckan, out_dir=out_dir, cancel_if_present=True, **kwargs)
        self._prepare_for_multithreading(ckan)
        self.read_line_counter = 0
        total = self.get_file_query_count()
        self.progress_callback.start_task(total, file_count=total, context=f"multi-thread download", level=CkanCallbackLevel.ResourceChunks)
        try:
            with ThreadPoolExecutor(max_workers=threads, initializer=self._init_thread, initargs=(ckan,)) as executor:
                if ckan.params.verbose_extra:
                    print(f"Launching multi-threaded download of multi-file resource {self.name}")
                futures = [executor.submit(self.download_file_query_item_graceful, ckan=ckan, out_dir=out_dir, file_query_item=file_chunk,
                                           index=index, external_stop_event=external_stop_event, start_index=start_index, end_index=end_index, **kwargs)
                           for index, file_chunk in enumerate(self.get_file_query_generator())]
                for future in futures:
                    future.result()  # This will propagate the exception
            self.progress_callback.end_task(total, file_count=total, context=f"multi-thread download", level=CkanCallbackLevel.ResourceChunks)
        except Exception as e:
            self.stop_event.set()  # Ensure all threads stop
            if ckan.params.verbose_extra:
                print(f"Stopping all threads because an exception occurred: {e}")
            raise e from e
        finally:
            # self.stop_event.set()  # Ensure all threads stop
            if ckan.params.verbose_extra:
                print("End of multi-threaded download...")
        # at last, apply final actions:
        self._terminate_thread()



#!python3
# -*- coding: utf-8 -*-
"""
Central implementation of lazy imports of optional dependencies / dependencies rarely used
"""
from types import SimpleNamespace

pyproj = None
def lazy_import_pyproj():
    global pyproj
    try:
        import pyproj
    except ImportError:
        pyproj = None
    return pyproj

shapely = SimpleNamespace(Geometry=None)
def lazy_import_shapely():
    global shapely
    try:
        import shapely
    except ImportError:
        shapely = SimpleNamespace(Geometry=None)
    return shapely

gpd = SimpleNamespace(GeoDataFrame=None)
def lazy_import_geopandas_gpd():
    global gpd
    try:
        import geopandas as gpd
    except ImportError:
        gpd = SimpleNamespace(GeoDataFrame=None)
    return gpd

bson = SimpleNamespace(ObjectId=None, DBRef=None)
def lazy_import_bson():
    global bson
    try:
        import bson
    except ImportError:
        bson = SimpleNamespace(ObjectId=None, DBRef=None)
    return bson

pymongo = SimpleNamespace(MongoClient=None, client_session=SimpleNamespace(ClientSession=None),
                          database=SimpleNamespace(Database=None), collection=SimpleNamespace(Collection=None))
def lazy_import_pymongo():
    global pymongo
    try:
        import pymongo
        import pymongo.client_session
        import pymongo.database
    except ImportError:
        pymongo = SimpleNamespace(MongoClient=None, client_session=SimpleNamespace(ClientSession=None),
                                  database=SimpleNamespace(Database=None), collection=SimpleNamespace(Collection=None))
    return pymongo

sqlalchemy = SimpleNamespace(Engine=None, Connection=None)
def lazy_import_sqlalchemy():
    global sqlalchemy
    try:
        import sqlalchemy
    except ImportError:
        sqlalchemy = SimpleNamespace(Engine=None, Connection=None)
    return sqlalchemy

psycopg2 = None
def lazy_import_psycopg2():
    global psycopg2
    try:
        import psycopg2
    except ImportError:
        psycopg2 = None
    return psycopg2

SSHTunnelForwarder = None
def lazy_import_ssh_tunnel_SSHTunnelForwarder():
    global SSHTunnelForwarder
    try:
        from sshtunnel import SSHTunnelForwarder
    except ImportError:
        SSHTunnelForwarder = None
    return SSHTunnelForwarder


#!python3
# -*- coding: utf-8 -*-
"""
Code to define the bondage between a file and a database query
in the context of a large DataStore defined by the concatenation of multiple files.
"""
from warnings import warn
from abc import ABC, abstractmethod
from typing import Dict, List, Iterable, Callable, Any, Tuple, Generator, Set, Union

import numpy as np
import pandas as pd

from ckanapi_harvesters.auxiliary.ckan_progress_callbacks import CkanProgressCallbackABC
from ckanapi_harvesters.builder.mapper_datastore import DataSchemeConversion
from ckanapi_harvesters.auxiliary.ckan_defs import ckan_tags_sep
from ckanapi_harvesters.auxiliary.ckan_model import CkanField
from ckanapi_harvesters.auxiliary.ckan_configuration import datastore_default_upload_index_col_name
from ckanapi_harvesters.ckan_api import CkanApi


class RequestMapperABC(DataSchemeConversion, ABC):
    """
    Class to define how to reconstruct a file from the full dataset
    This class overloads some data scheme conversion class functions
    This abstract class can be derived to specify custom data treatments
    """
    def __init__(self,
                 *, df_upload_fun:Callable[[pd.DataFrame], Any] = None,
                 df_download_fun:Callable[[pd.DataFrame], Any] = None):
        super().__init__(df_upload_fun=df_upload_fun, df_download_fun=df_download_fun)
        self.upsert_only_missing_rows:bool = False

    ## upsert request preparation  ----------------
    def get_file_query_of_df(self, df_upload:pd.DataFrame) -> Union[dict,None]:
        """
        Return the dict of {field: value} combinations representing the arguments of the query to reconstruct a file

        :param df_upload: the DataFrame representing the file
        :return:
        """
        return None

    def last_inserted_row_request(self, ckan:CkanApi, resource_id:str, file_query:dict) -> Union[pd.DataFrame,None]:
        """
        Request in CKAN the last inserted row(s) corresponding to a given file_query

        :param ckan:
        :param resource_id:
        :param file_query: a dict of {field: value} combinations representing the arguments of the query to reconstruct a file
        :return: The last row(s) in the database or None (if no specific method was defined)
        """
        return None

    def last_inserted_index_request(self, ckan:CkanApi, resource_id:str, file_query:dict, df_upload:pd.DataFrame) -> Tuple[int, bool, int, Union[pd.DataFrame,None]]:
        """
        Knowing the data which needs to be uploaded, this function compares the last known row(s) to the dataframe
        and returns the index to restart the upload process.

        :param ckan:
        :param resource_id:
        :param file_query: a dict of {field: value} combinations representing the arguments of the query to reconstruct a file
        :param df_upload: the known data corresponding to the file_query to be sent
        :return: a tuple (i_restart, upload_needed, row_count, df_last_row):
         - i_restart: the last known index in the dataframe
         - upload_needed: a boolean indicating if an update is necessary
         - row_count: the number of rows corresponding to the file_query
         - df_last_row: the last found row in the dataframe
        """
        return 0, True, -1, None

    ## download preparation  ----------------
    @abstractmethod
    def download_file_query_list(self, ckan: CkanApi, resource_id: str) -> List[dict]:
        """
        Function to list the {key: value} combinations present in the CKAN datastore to reconstruct the file database before downloading.

        :param ckan:
        :param resource_id:
        :return: a list of query arguments defining each file
        """
        raise NotImplementedError()

    def download_file_query_generator(self, ckan: CkanApi, resource_id: str) -> Generator[dict, Any, None]:
        """
        Generator for download_file_query_list which can be customized

        :param ckan:
        :param resource_id:
        :return:
        """
        for file_query in self.download_file_query_list(ckan=ckan, resource_id=resource_id):
            yield file_query

    def download_file_query(self, ckan: CkanApi, resource_id: str, file_query:dict, *, progress_callback:CkanProgressCallbackABC) -> Generator[pd.DataFrame, Any, None]:
        return ckan.datastore_search_page_generator(resource_id=resource_id, **file_query, search_all=True, progress_callback=progress_callback)


class RequestFileMapperABC(RequestMapperABC, ABC):
    """
    Class to define how to reconstruct a file from the full dataset
    This abstract class is oriented to treating files in the file system
    """
    def __init__(self,
                 *, df_upload_fun:Callable[[pd.DataFrame], Any] = None,
                 df_download_fun:Callable[[pd.DataFrame], Any] = None):
        super().__init__(df_upload_fun=df_upload_fun, df_download_fun=df_download_fun)
        self.file_name_prefix:str = "table_"
        self.file_name_suffix:str = ".csv"
        self.file_name_function:Union[Callable[[dict], str], None] = None

    def get_file_name_of_query(self, file_query:dict) -> str:
        if self.file_name_function is None:
            file_filters_str = '_'.join([str(key)+'_'+str(value) for key,value in file_query.items()])
        else:
            file_filters_str = self.file_name_function(file_query)
        return f"{self.file_name_prefix}{file_filters_str}{self.file_name_suffix}"


class RequestFileMapperUser(RequestFileMapperABC):
    """
    Use this basic implementation if the file query list is provided by the user or if the builder is only used to upload files.
    """
    def __init__(self, file_query_list: Iterable[Tuple[str, dict]],
                 *, df_upload_fun:Callable[[pd.DataFrame], Any] = None,
                 df_download_fun:Callable[[pd.DataFrame], Any] = None):
        super().__init__(df_upload_fun=df_upload_fun, df_download_fun=df_download_fun)
        # file_query_list must be stored in the BuilderDataStoreMultiAbc instance

    def download_file_query_list(self, ckan: CkanApi, resource_id: str) -> List[dict]:
        raise RuntimeError("File query list is provided by user")


class RequestFileMapperLimit(RequestFileMapperABC):
    """
    In this implementation, a file is defined by a certain amount of rows
    """
    default_limit = 10000

    def __init__(self, limit:int=None,
                 *, df_upload_fun:Callable[[pd.DataFrame], Any] = None,
                 df_download_fun:Callable[[pd.DataFrame], Any] = None):
        super().__init__(df_upload_fun=df_upload_fun, df_download_fun=df_download_fun)
        if limit is None:
            limit = RequestFileMapperLimit.default_limit
        self.limit:int = limit

    ## download preparation  ----------------
    def get_file_name_of_query(self, file_query:dict) -> str:
        if self.file_name_function is None:
            # file_filters_str = str(file_query["offset"] // self.limit)
            file_filters_str = f'{file_query["offset"]}_{file_query["offset"]+self.limit-1}'
        else:
            file_filters_str = self.file_name_function(file_query)
        return f"{self.file_name_prefix}{file_filters_str}{self.file_name_suffix}"

    def download_file_query_list(self, ckan: CkanApi, resource_id: str) -> List[dict]:
        # get number of rows and return a list of [offset,limit] combinations
        row_count = ckan.datastore_search_row_count(resource_id)
        return [{"offset": self.limit*counter, "limit": self.limit} for counter in range(row_count // self.limit + 1)]

    def download_file_query(self, ckan: CkanApi, resource_id: str, file_query:dict, *, progress_callback:CkanProgressCallbackABC) -> Generator[pd.DataFrame, Any, None]:
        return ckan.datastore_search_page_generator(resource_id=resource_id, offset=file_query["offset"], limit=file_query["limit"],
                                                    search_all=True, progress_callback=progress_callback)


class RequestFileMapperIndexKeys(RequestFileMapperABC):
    """
    In this implementation, a file is defined by a combination of file_keys values
    It is optionally ordered by an index_keys which enables to restart a transfer when interrupted
    By default, the index_keys is the last field of the primary key
    and the file_keys are the fields preceding the index_keys in the primary key
    """
    last_rows_limit = 1
    def __init__(self, group_by_keys:List[str], sort_by_keys:List[str] = None,
                 *, df_upload_fun:Callable[[pd.DataFrame], Any] = None,
                 df_download_fun:Callable[[pd.DataFrame], Any] = None):
        super().__init__(df_upload_fun=df_upload_fun, df_download_fun=df_download_fun)
        self.group_by_keys: List[str] = group_by_keys        # fields to filter to obtain one file
        self.sort_by_keys: Union[List[str],None] = None      # field to order the document
        if sort_by_keys is not None:
            self.sort_by_keys = sort_by_keys
        self.upload_add_index_column = ""  # override

    def get_necessary_fields(self) -> Set[str]:
        fields = set(self.group_by_keys)
        if self.sort_by_keys is not None:
            fields = fields.union(set(self.sort_by_keys))
        return fields

    def df_upload_alter(self, df_local: Union[pd.DataFrame, List[dict], Any], *, total_lines_read: int,
                        fields: Dict[str, CkanField], file_query: str, mapper_kwargs: dict = None, **kwargs) -> pd.DataFrame:
        # overload of df_upload_alter calling self.df_upload_fun
        # order dataframes before sending to database in order to be able to restart transfer from last transmitted index
        df_database = super().df_upload_alter(df_local, total_lines_read=total_lines_read, fields=fields,
                                              file_query=file_query, mapper_kwargs=mapper_kwargs, **kwargs)
        if self.sort_by_keys is not None:
            if self.df_upload_fun is None:
                df_database = df_database.copy()
            df_database.sort_values(self.sort_by_keys, inplace=True)
        return df_database

    ## upsert request preparation  ----------------
    def get_file_query_of_df(self, df_upload:pd.DataFrame) -> Union[dict,None]:
        df_file_query = df_upload[self.group_by_keys].drop_duplicates(subset=self.group_by_keys)
        if len(df_file_query) == 1:
            return {"filters": df_file_query.to_dict(orient="records")[0]}
        else:
            return None

    def last_inserted_row_request(self, ckan:CkanApi, resource_id:str, file_query:dict) -> Union[pd.DataFrame,None]:
        if self.sort_by_keys is None or not self.upsert_only_missing_rows:
            return None
        else:
            df = ckan.datastore_search(resource_id, filters=file_query["filters"], sort=ckan_tags_sep.join(self.sort_by_keys) + " desc",
                                       limit=self.last_rows_limit, search_all=False)  #, fields=self.file_keys + self.index_keys)
            return df

    def last_inserted_index_request(self, ckan:CkanApi, resource_id:str, file_query:dict, df_upload:pd.DataFrame) -> Tuple[int, bool, int, pd.DataFrame]:
        # df_upload is in the database format (df_upload_fun has been applied)
        # df_last_row has just been downloaded but no field typing has been applied
        df_last_row = self.last_inserted_row_request(ckan=ckan, resource_id=resource_id, file_query=file_query)
        if df_last_row is None or df_last_row.empty:
            return 0, True, df_last_row.attrs["total"] if df_last_row is not None else 0, df_last_row
        else:
            for key in self.sort_by_keys:
                if key in df_upload.columns:
                    # apply field typing from df_upload in order to perform line-by-line comparison
                    df_last_row[key] = df_last_row[key].astype(df_upload[key].dtype)
            match_table = np.column_stack([df_upload[key] == df_last_row[key].iloc[0] for key in self.sort_by_keys])
            match_array = np.logical_and.reduce(match_table, 1)
            i_restart = np.argwhere(match_array) + 1
            if len(i_restart) == 1 and len(i_restart[0]) == 1:
                i_restart_py = int(i_restart[0][0])
                return i_restart_py, i_restart_py < len(df_upload), df_last_row.attrs["total"], df_last_row
            elif len(i_restart) == 0:
                return 0, True, df_last_row.attrs["total"], df_last_row
            else:
                msg = "Multiple results obtained when querying the last inserted index"
                warn(msg)
                return 0, True, df_last_row.attrs["total"], df_last_row

    ## download preparation  ----------------
    def get_file_name_of_query(self, file_query:dict) -> str:
        if self.file_name_function is None:
            file_filters_str = '_'.join([str(key)+'_'+str(value) for key,value in file_query['filters'].items()])
        else:
            file_filters_str = self.file_name_function(file_query['filters'])
        return f"{self.file_name_prefix}{file_filters_str}{self.file_name_suffix}"

    def download_file_query_list(self, ckan: CkanApi, resource_id: str) -> List[dict]:
        # function to list the files which are defined by unique file_keys combinations in the database
        df_list = ckan.datastore_search(resource_id, fields=self.group_by_keys, search_all=True, distinct=True)
        # df_list = ckan.datastore_search(resource_id, filters={key: 0 for key in self.order_keys}, fields=self.file_keys, search_all=True)
        filters_list = df_list.to_dict(orient="records")
        return [{"filters": file_filter} for file_filter in filters_list]

#    def download_file_query(self, ckan: CkanApi, resource_id: str, file_query:dict) -> pd.DataFrame:
#        return ckan.datastore_search(resource_id=resource_id, filters=file_query["filters"], search_all=True)


def default_file_mapper_from_primary_key(primary_key:List[str]=None, file_query_list: Iterable[Tuple[str,dict]]=None) -> RequestFileMapperABC:
    if file_query_list is not None:
        mapper = RequestFileMapperUser(file_query_list)
    else:
        mapper = RequestFileMapperLimit()
    if primary_key is not None:
        if (len(primary_key) == 1 and datastore_default_upload_index_col_name is not None
                and primary_key[0].strip() == datastore_default_upload_index_col_name):
            # user can explicitly specify he wants to rely on the index created by default
            mapper.upload_index_column = datastore_default_upload_index_col_name
        elif len(primary_key) > 0:
            mapper.upload_index_column = ""  # do not create an extra index if the primary key is defined
    return mapper

